#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr  4 11:10:20 2019

@author: martindrech


"""

import numpy as np
from PyQt5 import QtCore
import threading
import time
from MCDAQcontrolsignals import anal_control_signal
from resources.messages import show_warning
from resources.scanFunctions import create_measurement_array, yield_array
from resources.storage import Storage


class CameraWorker(QtCore.QObject):
    imageReadySignal = QtCore.pyqtSignal(np.ndarray, int)
    measurementStartingSignal = QtCore.pyqtSignal(str)
    measurementEndingSignal = QtCore.pyqtSignal(str)


    def __init__(self, andor_camera, rigGen):
        super().__init__()

        self.cam = andor_camera
        self.rigGen = rigGen
        self.cam.Init()
        self.cam.SensorCooling.setValue(True)
        print("camera worker initialized")

        self.storage = Storage()
        self.ao_to_scan = None
        self.scan_array_gen = None
        self._lock = threading.Lock()
        self.measure_params = dict()

    def make_connections(self, frontend):
        # frontend connections
        live_acq_params = frontend.cameraParamFrame.paramTree.p
        frontend.cameraButton.clicked.connect(lambda: self.run(live_acq_params))
        frontend.signalWithMeasurementParameters.connect(self.simple_scan_measurement_start)
        frontend.roiDataReadySignal.connect(self.save_data_from_scan_step)
        frontend.measurementFrame.saveMeasureButton.clicked.connect(
                lambda: self.permanent_save_measurement(
                        comment=frontend.measurementFrame.comments.toPlainText()
                        )
                    )
                
        frontend.signalWithRigolMeasurementParameters.connect(self.rigol_measurement_start)
        frontend.rigolMeasurementFrame.saveMeasureButton.clicked.connect(
                lambda: self.permanent_save_measurement(
                        comment=frontend.rigolMeasurementFrame.comments.toPlainText()
                        )
                    )

        # internal connections
        self.cam.helper.imageAquiredSignal.connect(self.new_image_acquired)
        #self.cam.helper.timeoutSignal.connect()
    def __del__(self):
        print("adios camera worker")


    def run(self, params):
        if self.cam.CameraAcquiring.getValue():
            self._stop_acquisition_loop()
            print("acquisition now stopping")

        else:
            self._configure_acq_parameters(params)
            self._start_acquisition_loop()
            self.acq_thread = threading.Thread(
                target=self.cam.live_acquisition_loop
            )
            self.acq_thread.start()
            print("acquisiting now starting")

    def _start_acquisition_loop(self):
        self.cam.live_acquisition_configure()
        self.cam.AcquisitionStart()


    def _stop_acquisition_loop(self):
        self.acq_thread.do_run = False
        self.acq_thread.join()
        self.cam.AcquisitionStop()
        self.cam._flush()

    def _configure_acq_parameters(self, params):
        print('Acq parameters: ')
        for p in params.children()[0].children():
            print(p.name(), p.value())

        # setting info in parameter tree
        params.child('Basic acq parameters').child('Camera name').setValue(
            self.cam.CameraModel.getValue()
        )
        params.child('Basic acq parameters').child('Sensor Temperature').setValue(
            self.cam.SensorTemperature.getValue()
        )
        params.child('Basic acq parameters').child('Temperature Status').setValue(
            self.cam.TemperatureStatus.getString()
        )
        params.child('Basic acq parameters').child('Readout Time').setValue(
            self.cam.ReadoutTime.getValue()
        )
        # setting camera parameteres according to tree        
        self.cam.ExposureTime.setValue(
            params.child('Basic acq parameters').child('Exposure Time').value()
        )
        self.cam.FrameRate.setValue(self.cam.FrameRate.max())
        params.child('Basic acq parameters').child('Frame Rate').setValue(
            self.cam.FrameRate.getValue()
        )
        self.cam.AOIBinning.setString(
            params.child('Basic acq parameters').child('Pixel Binning').value()
        )
        self.cam.PixelReadoutRate.setString(
            params.child('Basic acq parameters').child('Pixel Readout Rate').value()
        )
        self.cam.SimplePreAmpGainControl.setString(
            params.child('Basic acq parameters').child('Dynamic Range').value()
        ) 
        self.cam.SpuriousNoiseFilter.setValue(
            params.child('Basic acq parameters').child('Spurious Noise Filter').value()=='True'
        )     
        
    @QtCore.pyqtSlot(int)
    def new_image_acquired(self, acq_index):
        self.imageReadySignal.emit(self.cam.acq_queue.get(), acq_index)
    
    def simple_scan_measurement_start(self, measure_params):

        if self.cam.CameraAcquiring.getValue():
            show_warning('Can not start measurement if camera is acquiring')
            return 0
        
        self.measurementStartingSignal.emit('simple')  
        
        self.ao_to_scan = self.get_scan_signal(measure_params['signal_to_scan'])
        self.ao_initial_value = self.ao_to_scan.AO.get_current_value()
        self.measure_params = measure_params
        start = self.measure_params['start']
        end = self.measure_params['end']
        step_num = self.measure_params['steps']
        self.scan_array_gen = yield_array(create_measurement_array(start, end, step_num))
        self.ao_to_scan.go_softly_to_value(start)
        
        time.sleep(3)
        
        self.storage.set_directory(measure_params['directory'])
        self.storage.create_data_file('Time', measure_params['signal_to_scan'], 'Roi data')
        self.storage.append_metadata_to_current_file(self.cam.GetMetaData(), self.measure_params)        
        
        self.cam.TriggerMode.setString('Software')

        self._start_acquisition_loop()        
        self.acq_thread = threading.Thread(
                target=self.cam.live_acquisition_loop
            )
        self.acq_thread.start()
        #with self._lock:
        self.simple_scan_measurement_step()
        
            
    def simple_scan_measurement_end(self):
        print('entering scan ending in worker')    
        self.measurementEndingSignal.emit('simple')  
        self.cam.trigger()
        self._stop_acquisition_loop()
        self.ao_to_scan.go_softly_to_value(self.ao_initial_value)     
        self.cam.TriggerMode.setString('Internal')
        print('exiting scan ending in worker')
        
    def simple_scan_measurement_step(self):
        try:
            print('Setting value of step')
            value = next(self.scan_array_gen)
            print('value set to ' + str(value))
            self.ao_to_scan.AO.set_out(value)
            print('value sent to ao')
            # aca puede esperar o chequear algo, por ahora nada
            self.cam.trigger()
            print('trigger')
        except StopIteration:
            self.simple_scan_measurement_end()
        except:
            raise
       
    def get_scan_signal(self, ao_name):
        return [ao for ao in anal_control_signal._registry if ao.name == ao_name][0]

    @QtCore.pyqtSlot(float, str)
    def save_data_from_scan_step(self, roiData, meas_type):
        if meas_type == 'simple':
            row = [time.time(), self.ao_to_scan.AO.current_value, roiData]
            self.storage.append_data_to_current_file(row)
            self.simple_scan_measurement_step()
        elif meas_type == 'rigol':
            row = [time.time(), float(self.rigGen.Frequency), roiData]
            self.storage.append_data_to_current_file(row)
            self.rigol_measurement_step()
        else:
            raise 'Oh, wrong meas type'

   
        
    
    def abort_measurement(self):
        self.scan_array_gen.close()
        print('Measurement aborted')
    
    @QtCore.pyqtSlot(str)
    def permanent_save_measurement(self, comment=''):
        
        self.storage.append_metadata_to_current_file(
                {'Final comment': comment}
                )
        self.storage.permanent_save_current_data()
    

    def rigol_measurement_start(self, measure_params):

        if self.cam.CameraAcquiring.getValue():
            show_warning('Can not start measurement if camera is acquiring')
            return 0
        
        self.measurementStartingSignal.emit('rigol')  
        
        self.measure_params = measure_params
        self.measure_params['Amplitude'] = self.rigGen.Voltage
        self.measure_params['Offset'] = self.rigGen.Offset
        
        start = self.measure_params['start']
        end = self.measure_params['end']
        step_num = self.measure_params['steps']
        self.scan_array_gen = yield_array(create_measurement_array(start, end, step_num))
        
        time.sleep(1)
        
        self.storage.set_directory(measure_params['directory'])
        self.storage.create_data_file('Time', 'frequency', 'Roi data')
        self.storage.append_metadata_to_current_file(self.cam.GetMetaData(), self.measure_params)        
        
        self.cam.TriggerMode.setString('Software')

        self._start_acquisition_loop()        
        self.acq_thread = threading.Thread(
                target=self.cam.live_acquisition_loop
            )
        self.acq_thread.start()
        #with self._lock:
        self.rigol_measurement_step()

    def rigol_measurement_step(self):
        try:
            print('Setting value of step')
            value = next(self.scan_array_gen)
            print('value set to ' + str(value))
            self.rigGen.Frequency = value
            print('value sent to rigol')
            # aca puede esperar o chequear algo, por ahora nada
            self.cam.trigger()
            print('trigger')
        except StopIteration:
            self.rigol_measurement_end()
        except:
            raise

    def rigol_measurement_end(self):
        print('entering scan ending in worker')    
        self.measurementEndingSignal.emit('rigol')  
        self.cam.trigger()
        self._stop_acquisition_loop()
        self.cam.TriggerMode.setString('Internal')
        print('exiting scan ending in worker')